from datetime import datetime, timedelta

from mobio.libs.m_scheduler_partitioning.scheduler_models.base_model import BaseModel


class SchedulerState:
    BUSY = "busy"
    FREE = "free"


class SchedulerStateModel(BaseModel):
    ID = "_id"
    STATE = "state"
    EXPIRY_TIME = "expiry_time"
    PARTITIONS = "partitions"
    ROOT_NODE = "root_node"

    def __init__(self, url_connection):
        super(SchedulerStateModel, self).__init__(url_connection)
        self.collection = "scheduler_state"

    def register_worker(self, worker_id, partitions, root_node, delay_time):
        raw_result = self.get_db().update(
            {
                self.ID: worker_id,
            },
            {
                "$set": {
                    self.ID: worker_id,
                    self.STATE: SchedulerState.FREE,
                    self.PARTITIONS: partitions,
                    self.ROOT_NODE: root_node,
                    self.EXPIRY_TIME: datetime.utcnow()
                    + timedelta(seconds=delay_time + 15),
                    self.UPDATED_TIME: datetime.utcnow(),
                },
                "$setOnInsert": {self.CREATED_TIME: datetime.utcnow()},
            },
            upsert=True,
        )
        upserted_id = (
            worker_id
            if raw_result.get("upserted")
            else worker_id
            if raw_result.get("updatedExisting")
            else None
        )
        return upserted_id

    def increase_expiry_time(self, worker_id, delay_time):
        return (
            self.get_db()
            .update_one(
                {self.ID: worker_id},
                {
                    "$set": {
                        self.STATE: SchedulerState.FREE,
                        self.EXPIRY_TIME: datetime.utcnow()
                        + timedelta(seconds=delay_time + 15),
                        self.UPDATED_TIME: datetime.utcnow(),
                    }
                },
            )
            .matched_count
        )

    def rebalance_partitions(self, worker_id, partitions):
        return (
            self.get_db()
            .update_one(
                {self.ID: worker_id},
                {
                    "$set": {
                        self.PARTITIONS: partitions,
                        self.UPDATED_TIME: datetime.utcnow(),
                    }
                },
            )
            .matched_count
        )

    def set_busy(self, worker_id):
        return (
            self.get_db()
            .update_one(
                {self.ID: worker_id},
                {
                    "$set": {
                        self.STATE: SchedulerState.BUSY,
                        self.UPDATED_TIME: datetime.utcnow(),
                        self.EXPIRY_TIME: datetime.utcnow() + timedelta(days=10),
                    }
                },
            )
            .matched_count
        )

    def get_free_worker(self, root_node):
        result = self.get_db().find(
            {self.ROOT_NODE: root_node, self.STATE: SchedulerState.FREE}
        )
        return list(result)



import os
from abc import abstractmethod
from time import sleep
from kazoo.client import KazooClient
from kazoo.recipe.watchers import DataWatch, ChildrenWatch
import ast
from mobio.libs.m_scheduler_partitioning import generate_nano_id
from mobio.libs.m_scheduler_partitioning.scheduler_models.scheduler_state_model import (
    SchedulerStateModel,
)


class MobioScheduler:
    rebalancing = False
    lst_partitions = []
    zk_client = None

    def register_worker(self):
        result_register_worker = None
        count_try = 0
        sleep_time = 3
        while not result_register_worker:
            if count_try >= 10:
                sleep_time = 10
            try:
                print("waiting for register worker")
                result_register_worker = SchedulerStateModel(self.url_connection).register_worker(
                    worker_id=self.node_id,
                    partitions=self.lst_partitions,
                    delay_time=self.delays,
                    root_node=self.root_node,
                )
            except Exception as ex:
                print("register_worker ERROR: {}".format(ex))
                result_register_worker = None
            print("result_register_worker: {}".format(result_register_worker))
            if not result_register_worker:
                sleep(sleep_time)
                count_try += 1
            else:
                break
        return result_register_worker

    def __init__(
        self,
        root_node="test-scheduler",
        node_id=None,
        nop=100,
        delays=1,
        url_connection=None,
    ):
        if not 10 <= nop <= 1000:
            raise Exception("nop {} not in range 10, 1000".format(nop))

        if delays > 3600:
            raise Exception("delays maximum is 3600 seconds")
        self.root_path = "/mobio-scheduler/{}".format(root_node)
        self.root_node = root_node
        self.nop = nop
        self.delays = delays
        self.url_connection = url_connection
        if not node_id:
            node_id = generate_nano_id(short=True)
        self.node_id = node_id
        self.my_node_path = "{}/{}".format(self.root_path, node_id)

        self.zk_client = KazooClient(
            hosts=os.getenv("ZOOKEEPER_CLUSTER", "localhost:2181")
        )
        self.zk_client.start()
        self.zk_client.ensure_path(self.root_path)
        sleep(1)
        self.zk_client.create(self.my_node_path, ephemeral=True)

        child_watch = ChildrenWatch(
            client=self.zk_client, path=self.root_path, func=self.watch_children
        )
        data_watch_root = DataWatch(
            client=self.zk_client, path=self.root_path, func=self.watch_data_root
        )
        data_watch_child = DataWatch(
            client=self.zk_client, path=self.my_node_path, func=self.watch_data_children
        )
        if self.url_connection:
            result_register_worker = self.register_worker()
            print("register worker success: {}".format(result_register_worker))
        try:
            while True:
                self.process()
                print("still alive after: {} seconds".format(self.delays))
                if self.url_connection:
                    result_increase_expiry_time = None
                    while not result_increase_expiry_time:
                        result_increase_expiry_time = SchedulerStateModel(self.url_connection).increase_expiry_time(
                            worker_id=self.node_id, delay_time=self.delays
                        )
                        print("result_increase_expiry_time: {}".format(result_increase_expiry_time))
                        if not result_increase_expiry_time:
                            result_register_worker = self.register_worker()
                            print("register worker success: {}".format(result_register_worker))
                sleep(self.delays)
        except RuntimeError as e:
            print("something unexpected happened: {}: {}".format(self.my_node_path, e))
        finally:
            print("consumer is stopped")
            self.zk_client.stop()
            self.zk_client.close()

    @abstractmethod
    def process(self):
        print("in process")
        if self.rebalancing:
            print("in process: rebalancing: {}".format(self.rebalancing))
            self.rebalancing = False
            print("in process: rebalancing: {}".format(self.rebalancing))
        """
        sample code:
        
        limit = 1000
        has_data = True
        sequence_id = 0
        lst_ids = []
        while has_data:
            lst_job = self.get_jobs(
                sequence_id=sequence_id, limit=limit
            )
            if len(lst_job) < limit:
                has_data = False
            for current_job in lst_job:
                if self.rebalancing:
                    if lst_ids:
                        self.save_to_sent_batch(lst_ids=lst_ids)
                        lst_ids = []
                        ConfluentProducerManager().flush_to_topic()
                    has_data = False
                    sys_conf.logger.warning(
                        "rebalancing in row: {}".format(current_msg.get("_id"))
                    )
                    self.rebalancing = False
                    break
    
                try:
                    sequence_id = current_job.get("sequence_id")
                    key = current_job.get("profile_id")
                    ConfluentProducerManager().send_message_to_topic_with_key_without_flush(
                        topic=KafkaTopic.FACTORY_F1_JB_ABANDONED_CART,
                        data=json.dumps(
                            json.loads(JSONEncoder().encode(current_job))
                        ).encode("utf-8"),
                        key=str(key),
                    )
    
                    lst_ids.append(_id)
                    if len(lst_ids) % 20 == 0:
                        self.save_to_sent_batch(lst_ids=lst_ids)
                        lst_ids = []
                        ConfluentProducerManager().flush_to_topic()
                except Exception as er:
                    err_msg = "do_process:: cannot sent data ERROR: {}".format(er)
                    sys_conf.logger.error(err_msg)
            if lst_ids:
                self.save_to_sent_batch(lst_ids=lst_ids)
                lst_ids = []
                ConfluentProducerManager().flush_to_topic()
        """

    # function theo dõi sự thay đổi của các workers.
    # Nếu có worker join hoặc leave parent thì sẽ assign lại partitions cho các workers
    def watch_children(self, lst_children):
        print("Children are now: {}".format(lst_children))
        self.rebalance_current_node()
        num_childs = len(lst_children)
        lst = [x for x in range(self.nop)]
        tmp_lst_partitions = self.split(lst, num_childs)
        for i in range(num_childs):
            value = bytes(str(tmp_lst_partitions[i]), "utf-8")
            self.zk_client.set(
                "{}/{}".format(self.root_path, lst_children[i]), value=value
            )
        print("set data done")

    # function notify số lượng partitions mà worker này sẽ xử lý
    def watch_data_root(self, data, stat):
        print("Version: %s, data: %s" % (stat.version, data.decode("utf-8")))

    def check_partition_in_parent(self, partition):
        lst_children = self.zk_client.get_children(path=self.root_path)
        result = True
        for children in lst_children:
            if children != self.node_id:
                children_data = ast.literal_eval(list(self.zk_client.get("{}/{}".format(self.root_path, children)))[0].decode("utf-8"))
                if partition in children_data:
                    print("ERROR:: node: {} children_data: {}, partition: {}".format(children, children_data, partition))
                    result = False
                    break
        return result

    # function notify cho worker là cần rebalance lại partitions
    def watch_data_children(self, data, stat):
        print(
            "watch_data_children:: Version: %s, data: %s"
            % (stat.version, data.decode("utf-8"))
        )
        d = data.decode("utf-8")
        self.rebalancing = True
        self.lst_partitions = []
        if d:
            array = ast.literal_eval(d)
            for p in array:
                result_check_partition_in_parent = self.check_partition_in_parent(partition=p)
                if result_check_partition_in_parent:
                    self.lst_partitions.append(p)
        print("current partition: {}".format(self.lst_partitions))
        if self.url_connection:
            result = None
            while not result:
                result = SchedulerStateModel(self.url_connection).rebalance_partitions(
                    worker_id=self.node_id, partitions=self.lst_partitions
                )
                print("result rebalance: {}".format(result))
                if not result:
                    result_register_worker = self.register_worker()
                    print("register worker success: {}".format(result_register_worker))

    # function reset số lượng partitions cúa worker này.
    def rebalance_current_node(self):
        self.zk_client.set(path=self.my_node_path, value=bytes("", "utf-8"))

    def split(self, a, n):
        k, m = divmod(len(a), n)
        return list(
            a[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n)
        )


if __name__ == "__main__":
    MobioScheduler()



# -*- coding: utf-8 -*-
from pymongo import MongoClient


def create_db(url_connection):
    print("create_db: ok")

    try:
        client = MongoClient(url_connection, connect=False)
    except Exception as ex:
        print('BaseModel::create_db: %r', ex)
        client = None

    return client


class DBManager:
    __instance = None

    @staticmethod
    def get_instance(url_connection):
        """ Static access method. """
        if DBManager.__instance is None:
            DBManager(url_connection)
        return DBManager.__instance

    def __init__(self, url_connection):
        """ Virtually private constructor. """
        if DBManager.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            print("DBManager: init call")
            DBManager.__instance = self
            self.db = create_db(url_connection)



import datetime
import json
import uuid

from bson import ObjectId

from mobio.libs.m_scheduler_partitioning.scheduler_models.db_manager import DBManager


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ObjectId):
            return str(o)
        elif type(o) == datetime.date:
            a = datetime.datetime.combine(o, datetime.datetime.min.time())
            return a.strftime("%Y-%m-%dT%H:%M:%S.000Z")
        elif type(o) == datetime.datetime:
            return o.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        elif isinstance(o, uuid.UUID):
            return str(o)
        elif isinstance(o, (bytes, bytearray)):
            return str(o, "utf-8")
        return json.JSONEncoder.default(self, o)


class BaseModel(object):
    CREATED_TIME = "created_time"
    UPDATED_TIME = "updated_time"

    db_name = 'test'
    url_connection = None
    collection = None

    db = None

    def __init__(self, url_connection):
        self.url_connection = url_connection
        self.client_mongo = DBManager.get_instance(self.url_connection).db
        self.db_name = self.client_mongo.get_database().name

    def get_db(self, read_preference=None):
        if not self.client_mongo:
            self.client_mongo = DBManager.get_instance(self.url_connection).db
        return self.client_mongo.get_database(self.db_name).get_collection(self.collection,
                                                                           read_preference=read_preference)

    @staticmethod
    def normalize_uuid(some_uuid):
        if isinstance(some_uuid, str):
            return uuid.UUID(some_uuid)
        return some_uuid

    @staticmethod
    def normalize_object_id(some_object_id):
        if isinstance(some_object_id, str):
            return ObjectId(some_object_id)
        return some_object_id



import os

from setuptools import setup
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


# cython detection
try:
    from Cython.Build import cythonize

    CYTHON = False
except ImportError:
    CYTHON = False

SOURCE_PATH = "./mobio"

ext_modules = []


# if CYTHON:
#     ext_modules = cythonize([SOURCE_PATH + "/**/*.py"], compiler_directives=dict(always_allow_keywords=True))


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


setup(
    name="m_scheduler_partitioning",
    version="0.1.5",
    description="Mobio scheduler partitioning sdk",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mobiovn",
    author="MOBIO",
    author_email="contact@mobio.vn",
    license="MIT",
    packages=[
        "mobio/libs/m_scheduler_partitioning",
        "mobio/libs/m_scheduler_partitioning/scheduler_models"
    ],
    install_requires=[
        "nanoid==2.0.0",
        "kazoo==2.8.0"
    ],
    # package_data={'': extra_files},
    include_package_data=True,
    ext_modules=ext_modules,
    classifiers=[
        "Topic :: Software Development",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)
